from hotsos.core.plugins.storage.ceph import CephChecksBase
from hotsos.core.utils import sorted_dict


class CephSummary(CephChecksBase):
    summary_part_index = 0

    def __0_summary_release(self):
        return {'name': self.release_name,
                'days-to-eol': self.days_to_eol}

    def __1_summary_services(self):
        """Get string info for running services."""
        if self.systemd.services:
            return self.systemd.summary
        if self.pebble.services:
            return self.pebble.summary

    def __2_summary_dpkg(self):
        # require at least one core package to be installed to include
        # this in the report.
        if self.apt.core:
            return self.apt.all_formatted

    def __2_summary_snaps(self):
        if self.snaps.core:
            return self.snaps.all_formatted

    def __3_summary_status(self):
        if self.cluster.health_status:
            return self.cluster.health_status

    def __4_summary_network(self):
        """ Identify ports used by Ceph daemons, include them in output
        for informational purposes.
        """
        net_info = {}
        for config, port in self.bind_interfaces.items():
            net_info[config] = port.to_dict()

        if net_info:
            return net_info

    def __5_summary_osd_pgs_near_limit(self):
        if self.cluster.osds_pgs_above_max:
            return self.cluster.osds_pgs_above_max

    def __6_summary_osd_pgs_suboptimal(self):
        if self.cluster.osds_pgs_suboptimal:
            return self.cluster.osds_pgs_suboptimal

    def __7_summary_versions(self):
        versions = self.cluster.ceph_daemon_versions_unique()
        if versions:
            return versions

    def __8_summary_mgr_modules(self):
        if self.cluster.mgr_modules:
            return self.cluster.mgr_modules

    def __9_summary_local_osds(self):
        if self.local_osds:
            osds = {}
            for osd in self.local_osds:
                osds.update(osd.to_dict())

            return sorted_dict(osds)

    def __10_summary_crush_rules(self):
        if self.cluster.crush_map.rules:
            return self.cluster.crush_map.rules

    def __11_summary_large_omap_pgs(self):
        if self.cluster.large_omap_pgs:
            return self.cluster.large_omap_pgs



from functools import cached_property

from propertree.propertree2 import PTreeLogicalGrouping
from hotsos.core.log import log
from hotsos.core.ycheck.engine.properties.common import (
    YPropertyOverrideBase,
    YPropertyMappedOverrideBase,
    YDefsSection,
    YDefsContext,
)
from hotsos.core.ycheck.engine.properties.requires.requires import (
    YPropertyRequires
)
from hotsos.core.ycheck.engine.properties.search import (
    YPropertySearch,
)
from hotsos.core.ycheck.engine.properties.inputdef import YPropertyInput


class CheckBase(object):

    def fetch_item_result(self, item):
        log.debug("%s: fetch_item_result() %s", self.__class__.__name__,
                  item.__class__.__name__)
        check = self.context['check']
        if isinstance(item, YPropertySearch):
            _results = check._search_results
            check._set_search_cache_info(_results)
            if not _results:
                log.debug("check %s search has no matches so result=False",
                          check.name)
                return False

            return True

        if isinstance(item, YPropertyRequires):
            result = item.result
            if result or check.cache.requires is None:
                check.cache.set('requires', item.cache)

            return result

        return item.result


class CheckLogicalGrouping(CheckBase, PTreeLogicalGrouping):
    _override_autoregister = False

    @property
    def result(self):
        try:
            return super().result
        except Exception as exc:
            log.exception("%s failed with exception: %s",
                          self.__class__.__name__, exc)
            raise


class YPropertyCheck(CheckBase, YPropertyMappedOverrideBase):
    _override_keys = ['check']
    _override_members = [YPropertyRequires, YPropertySearch, YPropertyInput]
    _override_logical_grouping_type = CheckLogicalGrouping

    @property
    def _search_results(self):
        """
        Retrieve the global searchkit.SearchResultsCollection from this
        property's context. We filter results using our tag and apply any
        search constraints requested.
        """
        global_results = self.context.search_results
        if global_results is None:
            raise Exception("no search results provided to check '{}'".
                            format(self.check_name))

        tag = self.search.unique_search_tag

        # first get simple search results
        simple_results = global_results.find_by_tag(tag)
        log.debug("check %s has %s simple search results with tag %s",
                  self.check_name, len(simple_results), tag)
        results = self.search.apply_extra_constraints(simple_results)

        seq_def = self.search.cache.sequence_search
        if not seq_def:
            return results

        # Not try for sequence search results
        sections = global_results.find_sequence_by_tag(tag).values()
        log.debug("check %s has %s sequence search results with tag %s",
                  self.check_name, len(sections), tag)
        if not sections:
            return results

        # Use the result of the first section start as the final
        # result. This is enough for now because the section is guaranteed
        # to be complete i.e. a full match and the result is ultimately
        # True/False based on whether or not a result was found.
        for result in list(sections)[0]:
            if result.tag == seq_def.start_tag:
                # NOTE: we don't yet support applying extra constraints here
                results.append(result)
                break

        return results

    @property
    def name(self):
        if hasattr(self, 'check_name'):
            return getattr(self, 'check_name')

    def _set_search_cache_info(self, results):
        """
        Set information in check cache so that it can be retrieved using
        PropertyCacheRefResolver. This information is typically used
        when creating messages as part of raising issues.

        IMPORTANT: do not cache search results themselves as this can consume a
                   lot of memory.

        @param results: search results for query in search property found in
                        this check.
        """
        # this is so that the information can be accessed like:
        # @checks.<checkname>.search.<setting>
        self.cache.set('search', self.search.cache)
        self.search.cache.set('num_results', len(results))
        if not results:
            return

        # Saves a list of files that contained search results.
        sources = set(r.source_id for r in results)
        files = [self.context.search_obj.resolve_source_id(s) for s in sources]
        self.search.cache.set('files', files)

    @cached_property
    def result(self):
        try:
            # Pass this object down to descendants so that they have access to
            # its cache etc. Note this is modifying global context but since
            # checks are processed sequentially this is fine.
            self.context['check'] = self
            log.debug("executing check %s", self.name)
            results = []
            stop_executon = False
            for member in self.members:
                for item in member:
                    # Ignore these here as they are used by search properties.
                    if isinstance(item, YPropertyInput):
                        continue

                    result = self.fetch_item_result(item)
                    results.append(result)
                    if CheckLogicalGrouping.is_exit_condition_met('and',
                                                                  result):
                        stop_executon = True
                        break

                if stop_executon:
                    break

            result = all(results)
            log.debug("check %s (%s) result=%s", self.name, results, result)
            return result
        except Exception:
            log.exception("something went wrong while executing check %s",
                          self.name)
            raise


class YPropertyChecks(YPropertyOverrideBase):
    _override_keys = ['checks']

    def initialise(self, local_vardefs, global_input, searcher, scenario):
        """
        Perform initialisation tasks for this set of checks.

        * create context containing vardefs for each check
        * pre-load searches from all/any checks and get results. This needs to
          be done before check results are consumed.

        @param local_vardefs: YPropertyVars object containing all variables
                              defined in the context of these checks and that
                              we wil pass on the all check def and properties.
        @param global_input: YPropertyInput object
        @param searcher: FileSearcher object
        """
        self.check_context = YDefsContext({'vars': local_vardefs})

        log.debug("pre-loading scenario '%s' checks searches into "
                  "filesearcher", scenario.name)
        # first load all the search definitions into the searcher
        for c in self._checks:
            if c.search:
                # local takes precedence over global
                _input = c.input or global_input
                if _input.command:
                    # don't apply constraints to command outputs
                    allow_constraints = False
                else:
                    allow_constraints = True

                for path in _input.paths:
                    log.debug("loading searches for check %s", c.check_name)
                    c.search.load_searcher(searcher, path,
                                           allow_constraints=allow_constraints)

        # provide results to each check object using global context
        self.check_context.search_obj = searcher

    @cached_property
    def _checks(self):
        log.debug("parsing checks section")
        if not hasattr(self, 'check_context'):
            raise Exception("checks not yet initialised")

        resolved = []
        for name, content in self.content.items():
            s = YDefsSection(self._override_name, {name: {'check': content}},
                             override_handlers=self.root.override_handlers,
                             resolve_path=self._override_path,
                             context=self.check_context)
            for c in s.leaf_sections:
                c.check.check_name = c.name
                resolved.append(c.check)

        return resolved

    def __iter__(self):
        log.debug("iterating over checks")
        for c in self._checks:
            log.debug("returning check %s", c.name)
            yield c



from hotsos.core import plugintools
from hotsos.core.config import HotSOSConfig
from hotsos.core.host_helpers import (
    APTPackageHelper,
    PebbleHelper,
    SystemdHelper,
)
from hotsos.core.ycheck.events import EventCallbackBase, EventHandlerBase
from hotsos.core.utils import sorted_dict

OVS_SERVICES_EXPRS = [r'ovsdb[a-zA-Z-]*',
                      r'ovs-vswitch[a-zA-Z-]*',
                      r'ovn[a-zA-Z-]*',
                      'openvswitch-switch',
                      ]
OVS_PKGS_CORE = ['openvswitch-switch',
                 'ovn',
                 ]
OVS_PKGS_DEPS = ['libc-bin',
                 'openvswitch',
                 'ovsdbapp',
                 'openssl',
                 'openvswitch-switch-dpdk',
                 ]
PY_CLIENT_PREFIX = r"python3?-{}\S*"


class OpenvSwitchChecksBase(plugintools.PluginPartBase):
    plugin_name = "openvswitch"
    plugin_root_index = 6

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        p_core = OVS_PKGS_CORE + [PY_CLIENT_PREFIX.format(p)
                                  for p in OVS_PKGS_CORE]
        p_deps = OVS_PKGS_DEPS + [PY_CLIENT_PREFIX.format(p)
                                  for p in OVS_PKGS_DEPS]
        self.apt = APTPackageHelper(core_pkgs=p_core, other_pkgs=p_deps)
        self.pebble = PebbleHelper(service_exprs=OVS_SERVICES_EXPRS)
        self.systemd = SystemdHelper(service_exprs=OVS_SERVICES_EXPRS)

    @property
    def plugin_runnable(self):
        # require at least one core package to be installed to run this plugin.
        return len(self.apt.core) > 0


class OpenvSwitchEventCallbackBase(EventCallbackBase):

    def categorise_events(self, *args, **kwargs):
        if 'include_time' not in kwargs:
            include_time = HotSOSConfig.event_tally_granularity == 'time'
            kwargs['include_time'] = include_time

        return super().categorise_events(*args, **kwargs)


class OpenvSwitchEventHandlerBase(OpenvSwitchChecksBase, EventHandlerBase):

    @property
    def summary(self):
        # mainline all results into summary root
        ret = self.run()
        if ret:
            return sorted_dict(ret)



from functools import cached_property

from hotsos.core.log import log
from hotsos.core.host_helpers import (
    APTPackageHelper,
    DPKGVersion,
)
from hotsos.core.ycheck.engine.properties.requires import (
    intercept_exception,
    YRequirementTypeBase,
    PackageCheckItemsBase,
)


class APTCheckItems(PackageCheckItemsBase):

    @cached_property
    def packaging_helper(self):
        return APTPackageHelper(self.packages_to_check)

    @cached_property
    def installed_versions(self):
        _versions = []
        for p in self.installed:
            _versions.append(self.packaging_helper.get_version(p))

        return _versions

    def normalize_version_criteria(self, version_criteria):
        """Normalize all the criterions in a criteria.

        Normalization does the following:
        - removes empty criteria
        - replaces old ops with the new ones
        - sorts each criterion(ascending) and criteria(descending)
        - adds upper/lower bounds to criteria, where needed

        @param version_criteria: List of version ranges to normalize
        @return: Normalized list of version ranges
        """

        # Step 0: Ensure that all version values are DPKGVersion type
        for idx, version_criterion in enumerate(version_criteria):
            for k, v in version_criterion.items():
                version_criterion.update({k: DPKGVersion(v)})

        # Step 1: Remove empty criteria
        version_criteria = [x for x in version_criteria if len(x) > 0]

        # Step 2: Replace legacy ops with the new ones
        legacy_ops = {"min": "ge", "max": "le"}
        for idx, version_criterion in enumerate(version_criteria):
            for lop, nop in legacy_ops.items():
                if lop in version_criterion:
                    version_criterion[nop] = version_criterion[lop]
                    del version_criterion[lop]

        # Step 3: Sort each criterion in itself, so the smallest version
        # appears first
        for idx, version_criterion in enumerate(version_criteria):
            version_criterion = dict(sorted(version_criterion.items(),
                                     key=lambda a: a[1]))
            version_criteria[idx] = version_criterion

        # Step 4: Sort all criteria by the first element in the criterion
        version_criteria = sorted(version_criteria,
                                  key=lambda a: list(a.values())[0])

        # Step 5: Add the implicit upper/lower bounds where needed
        lower_bound_ops = ["gt", "ge", "eq"]  # ops that define a lower bound
        upper_bound_ops = ["lt", "le", "eq"]  # ops that define an upper bound
        equal_compr_ops = ["eq", "ge", "le"]  # ops that compare for equality
        for idx, version_criterion in enumerate(version_criteria):
            log.debug("\tchecking criterion %s", str(version_criterion))

            has_lower_bound = any(x in lower_bound_ops
                                  for x in version_criterion)
            has_upper_bound = any(x in upper_bound_ops
                                  for x in version_criterion)
            is_the_last_item = idx == (len(version_criteria) - 1)
            is_the_first_item = idx == 0

            log.debug("\t\tcriterion %s has lower bound?"
                      "%s has upper bound? %s", str(version_criterion),
                      has_lower_bound, has_upper_bound)

            if not has_upper_bound and not is_the_last_item:
                op = "le"  # default
                next_criterion = version_criteria[idx + 1]
                next_op, next_val = list(next_criterion.items())[0]
                # If the next criterion op compares for equality, then the
                # implicit op added to this criterion should not compare for
                # equality.
                if next_op in equal_compr_ops:
                    op = "lt"
                log.debug("\t\tadding implicit upper bound %s:%s to %s", op,
                          next_val, version_criterion)
                version_criterion[op] = next_val
            elif not has_lower_bound and not is_the_first_item:
                op = "ge"  # default
                prev_criterion = version_criteria[idx - 1]
                prev_op, prev_val = list(prev_criterion.items())[-1]
                # If the previous criterion op compares for equality, then the
                # implicit op added to this criterion should not compare for
                # equality.
                if prev_op in equal_compr_ops:
                    op = "gt"
                log.debug("\t\tadding implicit lower bound %s:%s to %s", op,
                          prev_val, version_criterion)
                version_criterion[op] = prev_val

            # Re-sort and overwrite the criterion
            version_criteria[idx] = dict(
                sorted(version_criterion.items(),
                       key=lambda a: a[1]))

        # Step 6: Sort by descending order so the largest version range
        # appears first
        version_criteria = sorted(version_criteria,
                                  key=lambda a: list(a.values())[0],
                                  reverse=True)

        log.debug("final criteria: %s", str(version_criteria))
        return version_criteria

    def package_version_within_ranges(self, pkg, version_criteria):
        """Check if pkg's version satisfies any criterion listed in
        the version_criteria.

        @param pkg: The name of the apt package
        @param version_criteria: List of version ranges to normalize

        @return: True if ver(pkg) satisfies any criterion, false otherwise.
        """
        result = True
        pkg_version = self.packaging_helper.get_version(pkg)

        # Supported operations for defining version ranges
        ops = {
            "eq": lambda lhs, rhs: lhs == DPKGVersion(rhs),
            "lt": lambda lhs, rhs: lhs < DPKGVersion(rhs),
            "le": lambda lhs, rhs: lhs <= DPKGVersion(rhs),
            "gt": lambda lhs, rhs: lhs > DPKGVersion(rhs),
            "ge": lambda lhs, rhs: lhs >= DPKGVersion(rhs),
            "min": lambda lhs, rhs: ops["ge"](lhs, rhs),
            "max": lambda lhs, rhs: ops["le"](lhs, rhs),
        }

        version_criteria = self.normalize_version_criteria(version_criteria)

        for version_criterion in version_criteria:
            # Each criterion is evaluated on its own
            # so if any of the criteria is true, then
            # the check is also true.
            for op_name, op_fn in ops.items():
                if op_name in version_criterion:
                    version = str(version_criterion[op_name])
                    # Check if the criterion is satisfied or not
                    if not op_fn(str(pkg_version), version):
                        break
            else:
                # Loop is not exited by a break which means
                # all ops in the criterion are satisfied.
                result = True
                # Break the outer loop
                break
            result = False

        log.debug("package %s=%s within version ranges %s (result=%s)", pkg,
                  pkg_version, version_criteria, result)
        return result


class YRequirementTypeAPT(YRequirementTypeBase):
    """ Provides logic to perform checks on APT packages. """
    _override_keys = ['apt']
    _overrride_autoregister = True

    @property
    @intercept_exception
    def _result(self):
        _result = True
        items = APTCheckItems(self.content)

        # bail on first fail i.e. if any not installed
        if not items.not_installed:
            for pkg, versions in items:
                log.debug("package %s installed=%s", pkg, _result)
                if not versions:
                    continue
                _result = items.package_version_within_ranges(pkg, versions)
                # bail at first failure
                if not _result:
                    break
        else:
            log.debug("one or more packages not installed so returning False "
                      "- %s", ', '.join(items.not_installed))
            # bail on first fail i.e. if any not installed
            _result = False

        self.cache.set('package', ', '.join(items.installed))
        self.cache.set('version', ', '.join([
            str(x) for x in items.installed_versions]))
        log.debug('requirement check: apt %s (result=%s)',
                  ', '.join(items.packages_to_check), _result)
        return _result



import abc
from functools import cached_property
from collections import UserDict

from hotsos.core.config import HotSOSConfig
from hotsos.core.log import log
from hotsos.core.search import (
    FileSearcher,
    SearchConstraintSearchSince,
)
from hotsos.core.utils import sorted_dict
from hotsos.core.ycheck.engine import (
    YDefsLoader,
    YHandlerBase,
    YDefsSection,
)
from hotsos.core.ycheck.engine.properties import search
from hotsos.core.ycheck.engine.properties.search import CommonTimestampMatcher


CALLBACKS = {}


class EventCallbackNameConflict(Exception):
    pass


class EventCallbackNotFound(Exception):
    pass


class EventsSearchRegistryKeyNotFound(Exception):
    def __init__(self, key, all_keys):
        self.key = key
        self.all_keys = all_keys

    def __str__(self):
        return ("'{}' not found in event registry. Available keys are:"
                "\n      - {}".
                format(self.key, '\n      - '.join(self.all_keys)))


class EventCallbackMeta(type):

    def __init__(cls, _name, _mro, members):
        event_group = members.get('event_group')
        if event_group is None:
            return

        for event in members['event_names']:
            event = '{}.{}'.format(event_group, event)
            if event in CALLBACKS:
                msg = "event callback already registered: {}".format(event)
                raise EventCallbackNameConflict(msg)

            CALLBACKS[event] = cls


class EventCheckResult(object):
    """ This is passed to an event check callback when matches are found """

    def __init__(self, defs_section, defs_event, search_results, search_tag,
                 searcher, sequence_def=None):
        """
        @param defs_section: section name from yaml
        @param defs_event: event label/name from yaml
        @param search_results: searchkit.SearchResultsCollection
        @param search_tag: unique tag used to identify the results
        @param searcher: global FileSearcher object
        @param sequence_def: if set the search results are from a
                            searchkit.SequenceSearchDef and are therefore
                            grouped as sections of results rather than a single
                            set of results.
        """
        self.section = defs_section
        self.name = defs_event
        self.search_tag = search_tag
        self.results = search_results
        self.searcher = searcher
        self.sequence_def = sequence_def


class EventProcessingUtils(object):

    @classmethod
    def _get_event_results(cls, event):
        """
        Return dicts with keys 'date', 'key' and 'value' extracted from event
        search results. The result can have between one and three groups to
        match these keys respectively. If only one group (date) this provide a
        tally per date. Two groups provides a tally per key per date etc.
        """
        _fail_count = 0
        # Use these to ensure we only print the message once to avoid spamming
        # the logs.
        msg_flag = True
        for r in event.results:
            # the search expression used much ensure that these are
            # available in order for this to work.
            if len(r) > 2:
                if not set([None, None, None]).intersection([r.get(1),
                                                             r.get(2),
                                                             r.get(3)]):
                    yield {'date': r.get(1), 'time': r.get(2), 'key': r.get(3)}
                else:
                    _fail_count += 1
            elif len(r) == 0:
                msg = ("result (tag={}) does not have enough groups "
                       "(min 1) to be categorised - aborting".
                       format(r.tag))
                raise Exception(msg)
            else:
                if msg_flag and len(r) < 2:
                    msg_flag = False
                    # results with just a date will have None for the key
                    log.debug("results with tag=%s have just one group which "
                              "is assumed to be a date so using key=None",
                              r.tag)

                yield {'date': r.get(1), 'key': r.get(2)}

        if _fail_count:
            log.info("%s results ignored from event %s due to insufficient "
                     "fields",
                     _fail_count, event.name)

    @classmethod
    def _get_tally(cls, result, info, key_by_date, include_time,
                   squash_if_none_keys):
        if key_by_date:
            key = result['date']
            value = result['key']
        else:
            key = result['key']
            value = result['date']

        if key not in info:
            info[key] = {}

        if value is None and squash_if_none_keys:
            if info[key] == {}:
                info[key] = 1
            else:
                info[key] += 1
        else:
            ts_time = result.get('time')
            if value not in info[key]:
                if ts_time is not None and include_time:
                    info[key][value] = {}
                else:
                    info[key][value] = 0

            if ts_time is not None and include_time:
                if ts_time not in info[key][value]:
                    info[key][value][ts_time] = 1
                else:
                    info[key][value][ts_time] += 1
            else:
                info[key][value] += 1

    @classmethod
    def _sort_results(cls, categorised_results, key_by_date, include_time):
        log.debug("sorting categorised results")
        # sort main dict keys
        categorised_results = sorted_dict(categorised_results,
                                          reverse=not key_by_date)

        if all([key_by_date, include_time]):
            return

        # Sort values if they are dicts.
        for key, value in categorised_results.items():
            # If not using date as key we need to sort the values so that they
            # will have the same order as if they were sorted by date key.
            if not key_by_date:
                categorised_results[key] = sorted_dict(value)
                continue

            # sort by value i.e. tally/count
            for key, value in categorised_results.items():
                if not isinstance(value, dict):
                    break

                categorised_results[key] = sorted_dict(value,
                                                       key=lambda e: e[1],
                                                       reverse=True)

        return categorised_results

    @classmethod
    def categorise_events(cls, event, results=None, key_by_date=True,
                          include_time=False, squash_if_none_keys=False,
                          max_results_per_date=None):
        """
        Provides a generic way to categorise events. The default is to group
        events by key which is typically some kind of resource id or event
        label, beneath which the events are grouped by date. If a time value is
        also available, events will be grouped by time beneath their date. It
        may sometimes be useful to group events by date at the top level which
        is also supported here.

        The date, time and key values are extracted from each event search
        result and  are expected to be at indexes 1, 2 and 3 respectively. If
        the search result only has two groups it is assumed that the second is
        "key" i.e. no time is given. If the results argument is provided, this
        will be used instead of event.results.

        @param event: EventCheckResult object
        @param results: optional list of results where each item is a dict with
                        keys 'date', 'key' and optionally 'time'.
        @param key_by_date: by default results are categorised by date but in
                            situations where a small number of event types
                            (keys) are spread across many dates/times it might
                            make sense to categorise by event type.
        @param include_time: If true events will be categorised by time beneath
                             date to provide an extra level of granularity.
        @param squash_if_none_keys: If true and any key is found to be None
                                    (perhaps because a regex pattern did not
                                    match properly) the results for each date
                                    will be squashed to a number/count of
                                    events.
        @param max_results_per_date: an integer such that if key_by_date is
                                     True this will pick the top N entries
                                     with the highest count.
        """
        if results is None:
            results = cls._get_event_results(event)

        categorised_results = {}
        for r in results:
            cls._get_tally(r, categorised_results, key_by_date, include_time,
                           squash_if_none_keys)

        if not categorised_results:
            return

        categorised_results = cls._sort_results(categorised_results,
                                                key_by_date, include_time)
        if not (key_by_date and max_results_per_date):
            return categorised_results

        shortened = {}
        for date, entries in categorised_results.items():
            if len(entries) <= max_results_per_date:
                shortened[date] = entries
                continue

            top_n = dict(sorted(entries.items(),
                                key=lambda e: e[1],
                                reverse=True)[:max_results_per_date])
            shortened[date] = {'total': len(entries),
                               "top{}".format(max_results_per_date): top_n}

        return shortened


class EventCallbackBase(EventProcessingUtils, metaclass=EventCallbackMeta):
    # All implementations must set these and event_group must match that used
    # in the handler.
    event_group = None
    event_names = []

    @abc.abstractmethod
    def __call__(self):
        """ Callback method. """


class EventsSearchRegistry(UserDict):
    """
    Maintains a set of event names - dot paths to events in yaml tree - that
    have been registered as having a search property, a global FileSearcher
    object and the results from running searches. This information is used
    to load searches from a set of events, run them and save their results for
    later retrieval. Search results are tagged with the names stored here.

    It might be the case that an event handler wants to use its own
    FileSearcher in which case this supports setting a _custom_searcher that
    is cleared when the global searcher is accessed.
    """

    def __init__(self):
        self._global_searcher = None
        self._custom_searcher = None
        self._global_searcher_results = None
        super().__init__()

    def __getitem__(self, key):
        try:
            return super().__getitem__(key)
        except KeyError:
            raise EventsSearchRegistryKeyNotFound(
                                                key,
                                                list(self.data)) from KeyError

    def get_global_searcher_results(self):
        if self._global_searcher is None:
            raise Exception("registry global searcher is None")

        if self._global_searcher_results is not None:
            log.debug("using cached global event search results")
            return self._global_searcher_results

        log.debug("fetching global event search results")
        self._global_searcher_results = self._global_searcher.run()
        return self._global_searcher_results

    def get_global_searcher(self, allow_create=False):
        if self._global_searcher:
            log.debug("using existing global searcher (%s)",
                      self._global_searcher)
            return self._global_searcher

        if not allow_create:
            raise Exception("global events searcher is not set but is "
                            "expected to be.")

        self._custom_searcher = None
        constraint = SearchConstraintSearchSince(
                                         ts_matcher_cls=CommonTimestampMatcher)
        searcher = FileSearcher(constraint=constraint)
        self._global_searcher = searcher
        self._global_searcher_results = None
        log.debug("creating new global searcher (%s)", searcher)
        return searcher

    def set_custom_searcher(self, searcher):
        self._custom_searcher = searcher

    @property
    def current_searcher(self):
        return self._custom_searcher or self.get_global_searcher()

    def _reset_searchers(self):
        self._global_searcher = None
        self._custom_searcher = None
        self._global_searcher_results = None

    def reset(self, create_new_global_searcher=False):
        log.info("resetting events global registry")
        self._reset_searchers()
        self.data = {}
        if create_new_global_searcher:
            self.get_global_searcher(allow_create=True)


class EventsBase(object):
    # IMPORTANT: this state is maintained at class level so that all
    # implementations can share it. It is therefore crucial that state is reset
    # before loading a new set of event searches.
    search_registry = EventsSearchRegistry()

    @staticmethod
    def meets_requirements(event):
        """
        If an event or group has a requirements property it must return True
        for the events to be executed.
        """
        if HotSOSConfig.force_mode:
            return True

        if event.requires and not event.requires.result:
            log.debug("event '%s' pre-requisites not met - "
                      "skipping", event.name)
            return False

        return True

    @staticmethod
    def skip_filtered_event(event_path):
        """
        Apply event filter if provided.
        """
        e_filter = HotSOSConfig.event_filter
        if e_filter and event_path != e_filter:
            log.info("skipping event %s (filter=%s)", event_path, e_filter)
            return True

        return False

    @staticmethod
    def get_defs(group=None):
        """
        Load the event definitions for the current plugin. By default all are
        loaded and if a group path is provided, only events that are part of
        that group are included.

        @param group: a group path can be provided to include events part of a
                      group.
        """
        log.debug("loading event defs (group=%s)", group)
        plugin_defs = YDefsLoader('events').plugin_defs
        if not plugin_defs:
            return {}

        if not group:
            return plugin_defs

        # Exclude events that are not part of the group.
        groups = group.split('.')
        for i, subgroup in enumerate(groups):
            if i == 0:
                plugin_defs = {subgroup: plugin_defs[subgroup]}
            else:
                prev = groups[i - 1]
                plugin_defs[prev] = {subgroup: plugin_defs[prev][subgroup]}

        return plugin_defs

    @staticmethod
    def _get_event_from_path(events, path):
        """
        Walk down path until we hit the event containing the
        search property. We skip root/plugin name at start and
        ".search" at the end.

        @param event: YDefsSection object representing the entire tree of
                      events.
        @param path: event search property resolve path.
        """
        event = None
        for branch in path.split('.')[1:-1]:
            if event is None:
                event = getattr(events, branch)
            else:
                event = getattr(event, branch)

        return event

    @classmethod
    def _load_event_search(cls, event, searcher):
        """ Load search information from event into searcher.

        @param event: YDefsSection event object
        @param searcher: FileSearcher object
        """
        allow_constraints = True
        if event.input.command:
            # don't apply constraints to command outputs
            allow_constraints = False

        for path in event.input.paths:
            log.debug("loading search for event %s (path=%s, tag=%s)",
                      event.resolve_path,
                      path, event.search.unique_search_tag)
            # Add to registry in case it is needed by handlers e.g. for
            # sequence lookups.
            cls.search_registry[event.resolve_path] = {'search':
                                                       event.search}
            event.search.load_searcher(
                                searcher, path,
                                allow_constraints=allow_constraints)

    @classmethod
    def load_searches(cls, group=None, searcher=None):
        """
        Find all events that have a search property and load their search into
        the global searcher. A custom searcher will be used instead if
        provided.

        @param group: a group path can be provided to filter a subset of
                      events.
        @param searcher: customer FileSearcher object to be used instead of the
                         global searcher.
        """
        if searcher is None:
            searcher = cls.search_registry.get_global_searcher()
            if len(searcher.catalog) > 0:
                raise Exception("global event searcher catalog is not empty "
                                "and must be reset before loading so as not "
                                "to include searches from a previous run.")

        log.debug("started loading event (group=%s) searches into searcher "
                  "(%s)", group, searcher)

        search_props = set()
        events = YDefsSection(HotSOSConfig.plugin_name,
                              cls.get_defs(group) or {})
        for prop in events.manager.properties.values():
            for item in prop:
                if not issubclass(item['cls'], search.YPropertySearch):
                    break

                search_props.add(item['path'])

        if len(search_props) == 0:
            log.debug("finished loading event searches but no search "
                      "properties found")
            return

        log.debug("loading searches for %s events", len(search_props))
        for event_search_prop_path in search_props:
            event = cls._get_event_from_path(events, event_search_prop_path)
            if cls.skip_filtered_event(event.resolve_path):
                log.debug("skipping event %s", event.resolve_path)
                continue

            cls._load_event_search(event, searcher)

        log.debug("finished loading event searches into searcher "
                  "(registry has %s items)", len(cls.search_registry))


class EventsPreloader(EventsBase):
    """
    Pre-load all searches used in event definitions into a global FileSearcher
    object and execute the search before running any event callbacks.
    """

    @classmethod
    def execute(cls):
        # Pre-load all event searches into a global event searcher
        cls.load_searches()
        # Run the searches so that results are ready when event handlers are
        # run.
        cls.search_registry.get_global_searcher_results()

    @classmethod
    def reset(cls):
        # Make sure we start with a clean registry
        cls.search_registry.reset(create_new_global_searcher=True)

    @classmethod
    def run(cls):
        cls.reset()
        cls.execute()


class EventHandlerBase(EventsBase, YHandlerBase, EventProcessingUtils):
    """
    Root name used to identify a group of event definitions. Once all the
    yaml definitions are loaded this defines the level below which events
    for this checker are expected to be found.
    """
    event_group = None

    def __init__(self, *args, searcher=None, **kwargs):
        """
        @param searcher: optional FileSearcher object. If not provided then the
                         global searcher will be used which is the recommended
                         approach so that all searches are aggregated into one
                         operation and therefore files only need to be searched
                         once.
        """
        super().__init__(*args, **kwargs)
        if searcher is None:
            log.debug("no searcher provided - using global searcher")
            searcher = self.search_registry.get_global_searcher()
            # If no searcher is provided it is assumed that the global searcher
            # already exists, is loaded with searches and they have been
            # executed. Unit tests however, should be resetting the registry
            # prior to each run and we will therefore need to load searches
            # each time which is why we do this here. This is therefore not
            # intended to be used outside of a test scenario.
            if len(searcher.catalog) == 0:
                log.info("global searcher catalog is empty so launching "
                         "pre-load of event searches for group '%s'",
                         self.event_group)
                # NOTE: this is not re-entrant safe and is only ever expected
                #       to be done from a unit test.
                self.load_searches(group=self.event_group)
        else:
            # If a searcher is provided we switch over but do not clear global
            # searcher.
            if self.search_registry._custom_searcher != searcher:
                self.search_registry.set_custom_searcher(searcher)

            self.load_searches(group=self.event_group, searcher=searcher)

        self._event_results = None

    @property
    def searcher(self):
        """
        Return the current searcher we are using. If custom searcher is no
        longer needed it is expected that it will have been cleared in the
        __init__ method.
        """
        return self.search_registry.current_searcher

    @cached_property
    def events(self):
        """ Load event definitions from yaml. """
        group = YDefsSection(HotSOSConfig.plugin_name,
                             self.get_defs(self.event_group) or {})
        log.debug("sections=%s, events=%s",
                  len(list(group.branch_sections)),
                  len(list(group.leaf_sections)))

        _events = {}
        for event in group.leaf_sections:
            if self.skip_filtered_event(event.resolve_path):
                continue

            if not self.meets_requirements(event):
                return {}

            log.debug("event: %s", event.name)
            log.debug("input: %s (command=%s)", event.input.paths,
                      event.input.command is not None)

            section_name = event.parent.name
            if section_name not in _events:
                _events[section_name] = {}

            _events[section_name][event.name] = event.resolve_path

        return _events

    @property
    def final_event_results(self):
        """ Cache of results in case run() is called again. """
        return self._event_results

    def _get_event_search_results(self, event_search, global_results):
        if event_search.passthrough_results:
            # this is for implementations that have their own means of
            # retrieving results.
            return global_results

        seq_def = event_search.sequence_search
        if seq_def:
            search_results = global_results.find_sequence_sections(seq_def)
            if search_results:
                return search_results.values()
        else:
            return global_results.find_by_tag(event_search.unique_search_tag)

    def run(self, results=None):
        """
        Process each event and call respective callback functions when results
        where found.

        @param results: If no results are provides we get them from the global
                        searcher. This is provided for the case where a custom
                        searcher is in use.
        """
        if results is None:
            results = self.search_registry.get_global_searcher_results()

        if self.final_event_results is not None:
            return self.final_event_results

        if not CALLBACKS:
            raise Exception("need to register at least one callback for "
                            "event handler.")

        log.debug("registered event callbacks:\n%s", '\n'.
                  join(CALLBACKS.keys()))
        info = {}
        for section_name, section in self.events.items():
            for event, fullname in section.items():
                event_search = self.search_registry[fullname]['search']
                search_results = self._get_event_search_results(event_search,
                                                                results)
                if not search_results:
                    log.debug("event %s (tag=%s) did not yield any results",
                              event, event_search.unique_search_tag)
                    continue

                # We want this to throw an exception if the callback is not
                # defined.
                callback_name = '{}.{}'.format(self.event_group, event)
                if callback_name not in CALLBACKS:
                    msg = ("no callback found for event {}".
                           format(callback_name))
                    raise EventCallbackNotFound(msg)

                callback = CALLBACKS[callback_name]
                seq_def = event_search.sequence_search
                event_result = EventCheckResult(section_name, event,
                                                search_results,
                                                event_search.unique_search_tag,
                                                self.searcher,
                                                sequence_def=seq_def)
                log.debug("executing event %s.%s callback '%s'",
                          event_result.section, event, callback_name)
                ret = callback()(event_result)
                if not ret:
                    continue

                # if the return is a tuple it is assumed to be of the form
                # (<output value>, <output key>) where <output key> is used to
                # override the output key for the result which defaults to the
                # event name.
                if isinstance(ret, tuple):
                    out_key = ret[1]
                    ret = ret[0]
                    if not ret:
                        continue
                else:
                    out_key = event

                # Don't clobber results, instead allow them to aggregate.
                if out_key in info:
                    info[out_key].update(ret)
                else:
                    info[out_key] = ret

        if info:
            self._event_results = info
            return info



import re
from functools import cached_property

from hotsos.core.factory import FactoryBase
from hotsos.core.log import log
from hotsos.core.host_helpers import (
    CLIHelper,
    CLIHelperFile,
    HostNetworkingHelper
)
from hotsos.core.search import FileSearcher, SearchDef


class OVSDBTable(object):
    """
    Provides an interface to an OVSDB table. Records can be extracted from
    either 'get' or 'list' command outputs. We try 'get' first and of not found
    we search in output of 'list'.
    """

    def __init__(self, name):
        self.name = name

    def _convert_record_to_dict(self, record):
        """ Convert the ovsdb record dict format to a python dictionary. """
        out = {}
        if not record or record == '{}':
            return out

        expr = r'(\S+="[^"]+"|\S+=\S+),? ?'
        for field in re.compile(expr).findall(record):
            for char in [',', '}', '{']:
                field = field.strip(char)

            key, _, val = field.partition('=')
            out[key] = val.strip('"')

        return out

    def _get_cmd(self, table):
        return lambda **kwargs: CLIHelper().ovs_vsctl_get(table=table,
                                                          **kwargs)

    def _list_cmd(self, table):
        return lambda **kwargs: CLIHelper().ovs_vsctl_list(table=table,
                                                           **kwargs)

    def _fallback_query(self, column):
        """ Find first occurrence of column and return it. """
        for cmd in [self._list_cmd(self.name),
                    self._list_cmd(self.name.lower())]:
            for line in cmd():
                if line.startswith('{} '.format(column)):
                    return line.partition(':')[2].strip()

    def get(self, record, column):
        """
        Try to get column using get command and failing that, try getting it
        from list.
        """
        value = self._get_cmd(self.name)(record=record, column=column)
        if not value:
            value = self._fallback_query(column=column)

        return self._convert_record_to_dict(value)

    def __getattr__(self, column):
        """ Get column for special records i.e. with key '.' """
        return self.get(record='.', column=column)


class OVSDB(FactoryBase):
    """
    This class is used like a factory in that attributes are table names that
    return OVSDBTable objects on which attributes are row values.
    """
    def __getattr__(self, table):
        return OVSDBTable(table)


class OVSDPLookups(object):

    def __init__(self):
        cli = CLIHelper()
        out = cli.ovs_appctl(command='dpctl/show', flags='-s',
                             args='system@ovs-system')
        cexpr = re.compile(r'\s*lookups: hit:(\S+) missed:(\S+) lost:(\S+)')
        self.fields = {'hit': 0, 'missed': 0, 'lost': 0}
        for line in out:
            ret = re.match(cexpr, line)
            if ret:
                self.fields['hit'] = int(ret.group(1))
                self.fields['missed'] = int(ret.group(2))
                self.fields['lost'] = int(ret.group(3))

    def __getattr__(self, key):
        if key in self.fields:
            return self.fields[key]


class OVSBridge(object):

    def __init__(self, name, nethelper):
        self.name = name
        self.cli = CLIHelper()
        self.nethelper = nethelper

    @cached_property
    def ports(self):
        ports = []
        for line in self.cli.ovs_ofctl(command='show', args=self.name):
            ret = re.compile(r'^\s+\d+\((\S+)\):\s+').match(line)
            if ret:
                name = ret.group(1)
                port = self.nethelper.get_interface_with_name(name)
                if not port:
                    port = name

                ports.append(port)

        return ports


class OpenvSwitchBase(object):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.cli = CLIHelper()
        self.net_helper = HostNetworkingHelper()
        self.ovsdb = OVSDB()

    @cached_property
    def bridges(self):
        bridges = self.cli.ovs_vsctl_list_br()
        return [OVSBridge(br.strip(), self.net_helper) for br in bridges]

    @cached_property
    def tunnels(self):
        tunnel_info = {}
        ovn_external_ids = self.ovsdb.Open_vSwitch.external_ids
        if ovn_external_ids:
            # ovn only shows the local ip used in the db so we have to get from
            # there.
            proto = ovn_external_ids.get('ovn-encap-type')
            if proto:
                local_addr = ovn_external_ids['ovn-encap-ip']
                tunnel_info[proto] = {'local': local_addr}

        nethelp = HostNetworkingHelper()
        with CLIHelperFile() as cli:
            s = FileSearcher()
            expr = r'.+ \(([a-z]+): ([a-f\d\.:]+)->([a-f\d\.:]+), .+'
            s.add(SearchDef(expr, tag='all'),
                  cli.ovs_appctl(command='ofproto/list-tunnels'))
            results = s.run()
            for r in results.find_by_tag('all'):
                proto = r.get(1)
                if proto not in tunnel_info:
                    tunnel_info[proto] = {}

                if 'remotes' not in tunnel_info[proto]:
                    tunnel_info[proto]['remotes'] = []

                if 'local' not in tunnel_info[proto]:
                    tunnel_info[proto]['local'] = r.get(2)

                tunnel_info[proto]['remotes'].append(r.get(3))

        for proto, values in tunnel_info.items():
            local_addr = values['local']
            if local_addr:
                iface = nethelp.get_interface_with_addr(local_addr)
                if iface:
                    values['iface'] = iface.to_dict()
                    del values['local']
                else:
                    log.info("could not find interface for address '%s'",
                             local_addr)

            if 'remotes' not in values:
                log.info('no remote tunnel endpoints found for proto=%s',
                         proto)
                num_remotes = 0
            else:
                num_remotes = len(values['remotes'])

            values['remotes'] = num_remotes

        return tunnel_info

    @cached_property
    def offload_enabled(self):
        config = self.ovsdb.Open_vSwitch.other_config
        if not config:
            return False

        if config.get('hw-offload') == "true":
            return True

        return False


class OVSDPDK(OpenvSwitchBase):

    @cached_property
    def config(self):
        return self.ovsdb.Open_vSwitch.other_config or {}

    @property
    def enabled(self):
        if self.config.get('dpdk-init') == "true":
            return True

        return False

    @property
    def pmd_cpu_mask(self):
        mask = self.config.get('pmd-cpu-mask')
        if mask is not None:
            return int(mask, 16)

    @property
    def dpdk_lcore_mask(self):
        mask = self.config.get('dpdk-lcore-mask')
        if mask is not None:
            return int(mask, 16)



from .common import (  # noqa: F403,F401
    intercept_exception,
    OpsUtils,
    CheckItemsBase,
    YRequirementTypeBase,
    YRequirementTypeWithOpsBase,
    PackageCheckItemsBase,
    ServiceCheckItemsBase,
)



from hotsos.core.plugins.mysql import MySQLChecksBase


class MySQLSummary(MySQLChecksBase):
    summary_part_index = 0

    def __0_summary_services(self):
        if self.systemd.services:
            return self.systemd.summary
        if self.pebble.services:
            return self.pebble.summary

    def __1_summary_dpkg(self):
        if self.apt_info.core:
            return self.apt_info.all_formatted
